package Xray::FitParams;
require Exporter;

use strict;
use warnings;

use Xray::Graphics;

use Term::Complete;
use PDL;
use PDL::Graphics::PGPLOT;
use PDL::NiceSlice;

our $VERSION = 1.1;
our @ISA = ("Exporter");
our @EXPORT = qw(
  ParsInit
  PlotOne
  PlotAll
  PlotOneXY
  ReadOnePar
  BigPanel
  HalfPanel
  xyPanel
  xyPoints
  MinMaxAxis
  Descriptor
  SelMaxError
  $pg_parfile
  @pg_pars
  %pg_cols
  %pg_numeric
  $pg_npar
  $pg_nrows
  $pg_pars_verbouse
  $pg_symsize
  $pg_charsize
);  

our $pg_parfile;
our $pg_npar = 0;
our $pg_nrows = 0;
our @pg_pars = ();
our %pg_cols = ();
our %pg_numeric = ();
our $pg_pars_verbouse = 1;
our $pg_charsize = -1;
our $pg_symsize = -1;

####################################################

my $NOPE = 'nope';

sub ParsInit
{
  $pg_parfile = shift;

  $pg_npar = 0;
  $pg_nrows = 0;
  @pg_pars = ();
  %pg_cols = ();
  %pg_numeric = ();
  
  local *IN;   
  Descriptor();
  Numeric();
  open(IN, $pg_parfile) or die "\nAuch! Cannot open $pg_parfile\n";
  while(<IN>)
    {$pg_nrows++ if $_ !~ /^#/;}
  close(IN);
#  my $test = rcols $pg_parfile, $pg_cols{start};
#  $pg_nrows = at(nelem($test), 0);
  print "$pg_parfile: $pg_npar columns, $pg_nrows rows.\n\n" if($pg_pars_verbouse);
}

sub PlotOne
{
  my $par = shift;

  $pg_nx = 1;
  $pg_ny = 1;
  $pg_deltay = 0;
  $pg_ymin = 0.1;
  $pg_xmin = 0.1;
  
  PlotOneXY(1, 1, $par);
}

###

sub PlotAll
{
  $pg_nx = 1;
  $pg_ny = $pg_npar;
  $pg_deltay = 0;
  $pg_ymin = 0.07;
  $pg_xmin = 0.08;

  for my $i (0 .. $pg_npar - 1)
  {
    my $x = $i % $pg_nx + 1;
    my $y = int($i / $pg_nx) + 1;

    my $num = $i + 1;
    PlotOneXY($x, $y, $num);
  }
}

###

sub PlotOneXY
{
  my ($x, $y, $descr) = @_;
  
  my ($desc_x, $desc_y);
  if($descr =~ /:/)
    {($desc_x, $desc_y) = split /:/, $descr;}
  else
  {
    $desc_x = 'time';
    $desc_y = $descr;
  }
  
  my ($px, $px1, $px2, $namex, $logx, $maxxerr, $minxpar, $maxxpar) = ReadOnePar($desc_x);
  my ($py, $py1, $py2, $namey, $logy, $maxyerr, $minypar, $maxypar) = ReadOnePar($desc_y);
  my ($xmin, $xmax) = ($minxpar eq $NOPE) ? MinMaxAxis($px1, $px2, 0.1) : ($minxpar, $maxxpar);
  my ($ymin, $ymax) = ($minypar eq $NOPE) ? MinMaxAxis($py1, $py2, 0.1) : ($minypar, $maxypar); 

  my $sel = which(
    SelMaxError($px, $px1, $px2, $maxxerr) &
    SelMaxError($py, $py1, $py2, $maxyerr)
  );

  xyPanel($x, $y, $xmin, $xmax, $ymin, $ymax, $logx, $logy, $namex, $namey);
  xyPoints($px, $py, $px1, $px2, $py1, $py2, $sel, 'Black');
}
  
###

sub ReadOnePar
{
  my $desc = shift;
  $desc =~ tr/A-Z/a-z/ if($desc !~ /^F\d+/ && $desc !~ /^[SH]C/);  # ignore case
  
  my $maxerr = 1e6;
  my $minpar = $NOPE;
  my $maxpar = $NOPE;
  my $parname = '';
  my $log = 0;

  # range of a parameter e.g. /120,500/
  if($desc =~ s/\/([\deE\.\-\+]+),([\deE\.\-\+]+)\///)
  {
    $minpar = $1;
    $maxpar = $2;
  }
  
  # modifier norm.3 (maxerror) or norm.log (log scale)
  if($desc =~ /\./)
  {
    ($desc, my $modifier) = split /\./, $desc;
    if($modifier =~ s/log|lg|l//)
      {$log = 1;}
    if($modifier =~ /^\d+$/)
      {$maxerr = $modifier / 10;}
    elsif($modifier ne '')
      {die "Unrecognized modifier in $desc\n";}
  }
  
  # parameter can be a number or name
  if($desc =~ /^\d+$/)  # number
  {
    if($desc > $pg_npar) {die "Parameter number out of range: $desc\n";}
    $parname = $pg_pars[$desc - 1];
  }
  elsif($desc !~ /time/i)  # name but not time
  {
    if(!defined $pg_cols{$desc}) {die "Unidentified parameter in $pg_parfile: $desc\n";}
    $parname = $desc;
  }
  
  my ($par, $par1, $par2);
  if($desc =~ /^0$|time/i)
  {
    $parname = 'Time (MJD)';
    if(defined $pg_cols{start} && defined $pg_cols{stop})
    {
      ($par1, $par2) = rcols $pg_parfile, $pg_cols{start}, $pg_cols{stop};
      $par = ($par1 + $par2) / 2;
    }
    elsif(defined $pg_cols{time})
    {
      $par = rcols $pg_parfile, $pg_cols{time};
      $par1 = $par;
      $par2 = $par;
    }
    else
      {die "\nCannot find time column in the data\n";}
  }
  elsif($pg_numeric{$parname})
  {
    $par = rcols $pg_parfile, $pg_cols{$parname};
    $par1 = $par;
    $par2 = $par;
    # asymmetric error
    if(defined $pg_cols{"${parname}_1"})
      {($par1, $par2) = rcols $pg_parfile, $pg_cols{"${parname}_1"}, $pg_cols{"${parname}_2"};}
    # symmetric error
    if(defined $pg_cols{"${parname}_err"})
    {
      my $err = rcols $pg_parfile, $pg_cols{"${parname}_err"};
      $par1 = $par - $err;
      $par2 = $par + $err;
    }
    if($log)
    {
      $par = nlog10($par);
      $par1 = nlog10($par1);
      $par2 = nlog10($par2);
    }
  }
  else
  {
    $par = zeroes $pg_nrows;
    $par1 = $par;
    $par2 = $par;
  }
  return($par, $par1, $par2, $parname, $log, $maxerr, $minpar, $maxpar);
}


###

sub xyPanel
{
  my ($x, $y, $xmin, $xmax, $ymin, $ymax, $logx, $logy, $namex, $namey) = @_;
 
  my ($x1, $x2, $y1, $y2) = PanelCoord($x, $y);
  
  my $xtitle = ($y == $pg_ny) ? "$namex" : '';
  my $ytitle = ($x == 1) ? "$namey" : '';
  my $xaxis = ($y == $pg_ny) ? '1BCNST' : 'BCST';
  my $yaxis = ($x == 1) ? 'BCNST' : 'BCST';
  $xaxis .= 'L' if($logx);
  $yaxis .= 'L' if($logy);
  $pg_charsize = 1.1 - $pg_ny / 10 if($pg_charsize < 0);
  $pg_charsize = 0.6 if($pg_charsize < 0.6);
  env($xmin, $xmax, $ymin, $ymax, 
  {
    PlotPosition => [$x1, $x2, $y1, $y2], 
    Charsize => $pg_charsize, 
    Axis => [$xaxis, $yaxis],
    Xtitle => $xtitle, 
    Ytitle => $ytitle,
  });
#  text $lab, $left + 0.05*($right - $left), $top - 0.12*($top - $bottom), {Colour => 'Black', CharSize => 0.6};
}

###

sub BigPanel
{
  my ($xmin, $xmax, $ymin, $ymax, $logx, $logy, $namex, $namey) = @_;
 
  $pg_nx = 1;
  $pg_ny = 1;
  $pg_deltay = 0;
  $pg_ymin = 0.1;
  $pg_xmin = 0.1;

  xyPanel(1, 1, $xmin, $xmax, $ymin, $ymax, $logx, $logy, $namex, $namey);
}

###

sub HalfPanel
{
  my ($xmin, $xmax, $ymin, $ymax, $logx, $logy, $namex, $namey) = @_;
 
  $pg_nx = 1;
  $pg_ny = 2;
  $pg_deltay = 0;
  $pg_ymin = 0.1;
  $pg_xmin = 0.1;

  xyPanel(1, 2, $xmin, $xmax, $ymin, $ymax, $logx, $logy, $namex, $namey);
}


###

sub MinMaxAxis
{
  my ($x1, $x2, $expand) = @_;
  
  my $m = $x1->append($x2);
  my $min = $m->min();
  my $max = $m->max();

  my $delta = ($max > $min) ? $max - $min : $min;
  $delta = 1 if $delta == 0;
  my $xmin = $min - $delta * $expand;
  my $xmax = $max + $delta * $expand;
  
  return($xmin, $xmax);
}

###

sub xyPoints
{
  my ($x, $y, $x1, $x2, $y1, $y2, $sel, $colour, $symbol, $barcol) = @_;

  if(!defined $sel)
    {$sel = sequence $x;}
  if(!defined $colour)
    {$colour = 'Black';}
  if(!defined $symbol)
    {$symbol = $pg_sym{f_square};}
  if(!defined $barcol)#
    {$barcol = 'Gray';}
  
  my $x_l = $x - $x1;
  my $x_u = $x2 - $x;
  my $y_l = $y - $y1;
  my $y_u = $y2 - $y;
  
#  my $zero = zeroes($x->dice($sel));
  my $xx = $x->dice($sel);
  my $xx_l = $x_l->dice($sel);
  my $xx_u = $x_u->dice($sel);
  my $yy = $y->dice($sel);
  my $yy_l = $y_l->dice($sel);
  my $yy_u = $y_u->dice($sel);
  
  $pg_symsize = 1.1 - $pg_ny / 10 if($pg_symsize < 0);
  $pg_symsize = 0.6 if($pg_symsize < 0.6);

  errb $xx, $yy, $xx_l, $xx_u, $yy_l,  $yy_u, {Colour => $barcol, Symbol => 1, Charsize => 0.01, TERM => 1};
  points $xx, $yy, {Colour => $colour, Symbol => $symbol, Charsize => $pg_symsize};
}
  
###

sub Descriptor
{
  open(IN, $pg_parfile) or die "$!\n";
  my $line = <IN>;
  chomp $line;
#  $line =~ tr/A-Z/a-z/; # IGNORE CASE
  my @descr = split " ", $line;
  shift @descr;

  my @range = ();  # flux range
  my $col = 0;
  for my $d (@descr)
  {
    if($d =~ s/^F(\d+)\((.*)\)/F$1/) {push @range, $2;}   # read flux range
    elsif($d !~ /^[SH]C(_err)?$/)
      {$d =~ tr/A-Z/a-z/;}
    $pg_cols{$d} = $col++;
  }
  
#  shift @descr; shift @descr; shift @descr;
#  my $d;
#  do {
#    $d = shift @descr;
#    push @pg_pars;
#  } until ($d eq 'chired');
#  while($d = shift @descr)
#    {push @pg_pars, $d if($d =~ /^F\d+/ & $d !~ /_/);}

  while(my $d = shift @descr)
  {
    next if $d =~ /(name|start|stop|time)/;
    if($d =~ /(.*)_(1|2|err)$/)
      {next if defined $pg_cols{$1};}   # error of the parameter
    push @pg_pars, $d;
  }

  $pg_npar = @pg_pars;
  return if(!$pg_pars_verbouse);
  
  print "\n";
  my $num = 1;
  for my $p (@pg_pars)
  {
    print "  $num. $p \($pg_cols{$p}\)";
    print " with errors" if(defined $pg_cols{"${p}_1"} or defined $pg_cols{"${p}_err"});
    if($p =~ /^F\d+$/) {print " ", shift @range, " keV";}
    print "\n";
    $num++;
  }
  print "\n";
}


sub Numeric
{
  open(IN, $pg_parfile) or die "$!\n";
  my $line;
  while($line = <IN>)
    {last if $line !~ /^#/;}
  chomp $line;
  my @s = split " ", $line;
  for my $k (keys %pg_cols)
  {
    if($s[$pg_cols{$k}] =~ /^[+-]?((\d+)|(\d+\.\d*)|(\d*\.\d+))([eE][+-]?\d+)?$/)
      {$pg_numeric{$k} = 1;}
  }
}
  
####

sub SelMaxError
{
  my ($x, $x1, $x2, $maxerror) = @_;
  return((abs($x - $x1) <= $maxerror * abs($x)) & (abs($x - $x2) <= $maxerror * abs($x)));
}  



1;
