package Xray::Graphics;
require Exporter;

use strict;
use warnings;

use Term::Complete;
use PDL;
use PDL::Graphics::PGPLOT;
use PDL::NiceSlice;

our @ISA = qw(Exporter);
our @EXPORT = qw(
  AbsoluteText
  LogSerr
  LogErr
  nlog10
  PanelCoord
  Rebin
  iRebin
  $pg_nx
  $pg_ny
  $pg_xmin
  $pg_xmax
  $pg_ymin
  $pg_ymax
  $pg_deltax
  $pg_deltay
  %pg_sym
);  

our $pg_nx = 1;
our $pg_ny = 1;
our $pg_xmin = 0.05;
our $pg_xmax = 0.95;
our $pg_ymin = 0.05;
our $pg_ymax = 0.95;
our $pg_deltax = 0.01;
our $pg_deltay = 0.01;
our $pg_lwidth = 2;
our %pg_sym = ();

####################################################

my $symnum1 = 0;
my @sym1 = qw(square dot plus asterisk circle x square triangle earth sun psquare diamond star f_triangle cross david f_square f_circle f_star);

my $symnum2 = 627;
my @sym2 = qw(alpha beta gamma delta epsilon zeta eta theta iota kappa lambda mu nu xi omicron pi rho sigma tau upsilon phi chi psi omega);

my $symnum3 = 527;
my @sym3 = qw(Alpha Beta Gamma Delta Epsilon Zeta Eta Theta Iota Kappa Lambda Mu Nu Xi Omicron Pi Rho Sigma Tau Upsilon Phi Chi Psi Omega);

pg_mk_symbols(\@sym1, $symnum1);
pg_mk_chars(\@sym2, $symnum2);
pg_mk_chars(\@sym3, $symnum3);

$pg_sym{times} = '\\(0727)';
$pg_sym{odot} = '\\(2281)';

sub pg_mk_symbols
{
  my ($sym, $symnum) = @_;
  for my $name (@$sym)
    {$pg_sym{$name} = $symnum++}
}

sub pg_mk_chars
{
  my ($sym, $symnum) = @_;
  for my $name (@$sym)
  {
    $pg_sym{$name} = "\\($symnum)";
    $symnum++;
  }
}

####################################################

sub AbsoluteText
#
# AbsoluteText($text, $x, $y, [$colour, [$charsize]]);
#
# Places text in absolute coordinates of the picture (0-1).
#
{
  my ($text, $x, $y, $colour, $charsize) = @_;
  
  $colour = 'Black' unless defined $colour;
  $charsize = 1.3 unless defined $charsize;
  
  my $eps = 1e-5;
  env(0, 1, 0, 1, {PlotPosition => [$x, $x + $eps, $y, $y + $eps], Axis => 'Empty'});
  text $text, 0, 0, {Colour => $colour, Charsize => $charsize};
}

####################################################

sub LogSerr
#
# ($logx, $logx_lower, $logx_upper) = LogSerr($x, $dx);
#
# Calculates logs of $x, $x - $dx and $x + $dx (i.e.
# lower and upper error), replacing zero and negative values
# with $tiny (see nlog10).
#
{
  my ($x, $dx) = @_;
  
  my $x1 = $x - $dx;
  my $x2 = $x + $dx;
  
  my $lx = nlog10($x);
  my $err_l = $lx - nlog10($x1);
  my $err_u = nlog10($x2) - $lx;
  return($lx, $err_l, $err_u);
}

####################################################

sub LogErr
#
# ($logx, $logx_lower, $logx_upper) = LogSerr($x, $x1, $x2);
#
# Calculates logs of $x, $$x1 and $x2 (i.e.
# lower and upper error), replacing zero and negative values
# with $tiny (see nlog10).
#
{
  my ($x, $x1, $x2) = @_;
  
  my $lx = nlog10($x);
  my $err_l = $lx - nlog10($x1);
  my $err_u = nlog10($x2) - $lx;
  return($lx, $err_l, $err_u);
}

####################################################
sub nlog10
{
#
# $logx = nlog10($x);
#
# log of $x, zero and negative values are replaced by $tiny
#
  my $x = shift;
  my $tiny = 1e-32;
  
  my $m = $x > 0;
  $x = $x * $m + $tiny * (1 - $m);
  my $y = log10($x);
  return($y);
}

####################################################

sub PanelCoord
#
# ($x1, $x2, $y1, $y2) = PanelCoord($x, $y);
#
# Returns coordinates (to be used in PlotPosition inside env)
# of a panel ($x, $y). There are some global variables that
# can be changed:
#
# $pg_nx, $pg_ny - number of panels in each direction
# $pg_xmin, $pg_xmax, $pg_ymin, $pg_ymax - borders of
#    the entire plot (between 0 and 1)
# $pg_deltax, $pg_deltay - gaps between panels (in 0-1 units)
#
{
  my ($x, $y) = @_;
    
  my $sizex = ($pg_xmax - $pg_xmin) / $pg_nx;
  my $sizey = ($pg_ymax - $pg_ymin) / $pg_ny;
  my $x1 =  ($pg_xmin + ($x - 1) * $sizex);
  my $x2 = $x1 + $sizex - $pg_deltax;
  my $y1 = $pg_ymax - $y * $sizey;
  my $y2 = $y1 + $sizey - $pg_deltay;
  
  return($x1, $x2, $y1, $y2);
}
####################################################


sub Rebin
#
#  ($r_x, $r_dx, $r_y, $r_dy, $idx) = 
#     Rebin($x, $dx, $y, $dy, $accuracy, $maxnum, $maxgap)
#
#  Rebins spectrum/light curve. The bins are merged together
#  until the desired $accuracy = sigma_y/y is reached or
#  $maxnum number of points have been merged.
#
#  $maxgap is the maximum gap in the light curve (in $x units)
#  over which the points are merged. In addition, $maxgap controls
#  the way data points are merged. If defined (light curve case)
#  the $y values are weighted according to the length of each bin.
#  If $maxgap is not defined (spectrum case) the bins are weighted
#  according to their errors, $dy.
#
#  $idx is the rebin index that can be used to rebin other data
#  in the same way.
#
{
  my ($x, $dx, $y, $dy, $accuracy, $maxnum, $maxgap, $spectrum) = @_;
  
  my $huge = 1e36;
  
  if(!defined $maxgap)
  {
    $maxgap = $huge;
    $spectrum = 1 if !defined $spectrum;
  }
  $spectrum = 0 if !defined $spectrum;
 
  my ($sx, $sdx, $sy, $sdy) = (at($x, 0), at($dx, 0), at($y, 0), at($dy, 0));
  
  my $idx = zeroes $x;
  my $rx = zeroes $x;
  my $rdx = zeroes $dx;
  my $ry = zeroes $y;
  my $rdy = zeroes $dy;
  
  my $n = nelem($x);   
  my $num = 1;
  for my $i (1 .. $n - 1)
  {
    my $gap = $x($i) - $dx($i) - ($sx + $sdx);
    #if(($sdy > $sy * $accuracy || $sy <= 0) && ($dy($i) > $y($i) * $accuracy || $y($i) <= 0) && $num < $maxnum && $gap <= $maxgap)
    if(($sdy > $sy * $accuracy || $sy <= 0) && $num < $maxnum && $gap <= $maxgap)
    {
    # accumulate bins
      my $len = $sdx + $dx($i);
      my $interv = ($x($i) + $dx($i) - ($sx - $sdx))/2;
      $sx = ($sx - $sdx + $x($i) + $dx($i)) / 2;
      if(!$spectrum)  # light curve
      {
        $sy = ($sy * $sdx + $y($i) * $dx($i)) / $len;
        $sdy = sqrt($sdx**2 * $sdy**2 + $dx($i)**2 * $dy($i)**2) / $len;
      }
      else # spectrum
      {
        $sy = ($sy/$sdy**2 + $y($i)/$dy($i)**2) / (1/$sdy**2 + 1/$dy($i)**2);
        $sdy = sqrt(1 / (1/$sdy**2 + 1/$dy($i)**2));
      }
      $sdx = $interv;
      $num++;
    }
    else
    {
      # store accumulated bins
      $idx($i-1) .= 1;
      $rx($i-1) .= $sx;
      $rdx($i-1) .= $sdx;
      $ry($i-1) .= $sy;
      $rdy($i-1) .= $sdy;
      ($sx, $sdx, $sy, $sdy) = (at($x, $i), at($dx, $i), at($y, $i), at($dy, $i));
      $num = 1;
    }
  }
  $idx($n-1) .= 1; # always push the last one
  $rx($n-1) .= $sx;
  $rdx($n-1) .= $sdx;
  $ry($n-1) .= $sy;
  $rdy($n-1) .= $sdy;
  
  my $r_x = $rx->where($idx);  # collapse arrays
  my $r_dx = $rdx->where($idx);
  my $r_y = $ry->where($idx);
  my $r_dy = $rdy->where($idx);
  
  return($r_x, $r_dx, $r_y, $r_dy, $idx);
}

####################################################

sub iRebin
#
# ($r_x, $r_dx, $r_y, $r_dy) = iRebin($x, $dx, $y, $dy, $idx);
#
#  Rebins data according to the index built with Rebin.
#  Needs to be rewritten in piddle style.
#
{
  my ($x, $dx, $y, $dy, $idx) = @_;
  
  
  my ($sx, $sdx, $sy, $sdy) = ($x(0), $dx(0), $y(0), $dy(0));
  my $n = nelem($x);

  my $rx = zeroes($x);
  my $rdx = zeroes($dx);
  my $ry = zeroes($y);
  my $rdy = zeroes($dy);
    
  for my $i (1 .. $n - 1)
  {
    if($idx($i-1) != 1)
    {
      my $len = $sdx + $dx($i);
      my $interv = ($x($i) + $dx($i) - ($sx - $sdx))/2;
      $sx = ($sx - $sdx + $x($i) + $dx($i)) / 2;
      $sy = ($sy * $sdx + $y($i) * $dx($i)) / $len;
      $sdy = sqrt($sdx**2 * $sdy**2 + $dx($i)**2 * $dy($i)**2) / $len;
      $sdx = $interv;
    }
    else
    {
      $rx($i-1) .= $sx;
      $rdx($i-1) .= $sdx;
      $ry($i-1) .= $sy;
      $rdy($i-1) .= $sdy;
      ($sx, $sdx, $sy, $sdy) = ($x($i), $dx($i), $y($i), $dy($i));
    }
  }

  $rx($n-1) .= $sx;
  $rdx($n-1) .= $sdx;
  $ry($n-1) .= $sy;
  $rdy($n-1) .= $sdy;
  
  my $newx = $rx->where($idx);
  my $newdx = $rdx->where($idx);
  my $newy = $ry->where($idx);
  my $newdy = $rdy->where($idx);
  return($newx, $newdx, $newy, $newdy);
}
  
###


1;
