package Xray::PDS;
require Exporter;

use strict;
use warnings;

use Math::Trig;
use Xray::RXTE;
use Xray::Tools;
use Xray::FITS;

our $VERSION = 1.00;

our @ISA = ("Exporter");
our @EXPORT = qw(
 BkgPDS
 DeadCurves
 DeadPDS
 RmsPds
 RmsPdsBkg
 RmsPds2
 StatLc
);

my $PI = 4 * atan(1);

################################################################

sub BkgPDS
#
# Correct PDS for background count rate
#
# BkgPDS(\@pds, $src, $bkg);
#
# @pds - PDS array containing ($f, $df, $p, $dp) bins.
#        This is a standard PDS created by powspec with
#        normalization = -2.
# $src - source average count rate (background subtracted)
# $bkg - background average count rate
#
{
  my ($pds, $src, $bkg) = @_;
  my $corr = (($src + $bkg) / $src)**2;
  
  for my $i (0 .. @$pds - 1)
  {
    $pds->[$i][2] *= $corr;
    $pds->[$i][3] *= $corr;
  }
}

####################################################

sub DeadCurves
#
# Creates nonVLE and VLE lightcurves from Standard 1 files
# and calculates their count rates, used for dead-time corrections
#
# Usage:
#
#   ($nonVLE_cnt, $VLE_cnt, $ontime, $stat) 
#     = DeadCurves($root, $obsid, \@std1, $gti, \@log, clean);
#
{
  my ($root, $obsid, $std1, $gti, $log, $clean) = @_;
  
  $clean = 1 if !defined $clean;
  
  my $tmp_std1_list = "/tmp/dead_std1_$$.lst";
  my %obsstd1 = ObsFileList(\@{$std1});

  local *TMP;
  open(TMP, ">$tmp_std1_list") or die "\nCannot create $tmp_std1_list\n";
  foreach my $f (@{$obsstd1{$obsid}})
  {
    print TMP "$f\n";
  }
  close(TMP);
 
  my $nonVLE = "${LC_DIR}/${root}_${obsid}_nonVLE";
  my $VLE = "${LC_DIR}/${root}_${obsid}_VLE";
  my $nonVLE_cols = '"XeCntPcu0 XeCntPcu1 XeCntPcu2 XeCntPcu3 XeCntPcu4 RemainingCnt VpCnt"';
  my $VLE_cols = 'VLECnt';

  my $stat = 0;
  my ($args, @tmplog, $dum);
  
  if(!(-e "${nonVLE}.lc"))
  {
    print "  Creating nonVLE lightcurve...";
    $args = "saextrct infile=\@$tmp_std1_list gtiorfile=APPLY gtiandfile=$gti outroot=$nonVLE extenlc=.lc timecol=TIME columns=$nonVLE_cols multiple=no binsz=0.125 printmode=LIGHTCURVE lcmode=RATE spmode=SUM timemin=INDEF timemax=INDEF timeint=INDEF chmin=INDEF chmax=INDEF chint=INDEF chbin=INDEF clobber=yes mode=h";
    @tmplog = &call($args, $stat);
    push @$log, @tmplog;
    print "\n";
    if($stat) {return (0, 0, 0, $stat);}
  }
  
  if(!(-e "${VLE}.lc"))
  {
    print "  Creating VLE lightcurve...";
    $args = "saextrct infile=\@$tmp_std1_list gtiorfile=APPLY gtiandfile=$gti outroot=$VLE extenlc=.lc timecol=TIME columns=$VLE_cols multiple=no binsz=0.125 printmode=LIGHTCURVE lcmode=RATE spmode=SUM timemin=INDEF timemax=INDEF timeint=INDEF chmin=INDEF chmax=INDEF chint=INDEF chbin=INDEF clobber=yes mode=h";
    @tmplog = &call($args, $stat);
    push @$log, @tmplog;
    print "\n";
    if($stat) {return (0, 0, 0, $stat);}    
  }

  if(!-e "${nonVLE}.lc" || !-e "${VLE}.lc")
    {return (0, 0, 0, $stat);}    
    
  my $key = `fkeyprint infile=${nonVLE}.lc keynam=ONTIME outfile="STDOUT" | grep =`;
  $key =~ /=\s+(.*) \//;
  my $ontime = $1;

  my ($nonVLE_cnt, $VLE_cnt);
  my @fstat = `fstatistic infile=${nonVLE}.lc colname=RATE rows=-`;
  ($dum, $nonVLE_cnt) = split(/is/, $fstat[1]);
  chomp $nonVLE_cnt;

  @fstat = `fstatistic infile=${VLE}.lc colname=RATE rows=-`;
  ($dum, $VLE_cnt) = split(/is/, $fstat[1]);
  chomp $VLE_cnt;
  
  unlink "${nonVLE}.lc", "${VLE}.lc" if $clean;

  return($nonVLE_cnt, $VLE_cnt, $ontime, $stat);
}

###################################################################

sub DeadPDS
#
# Correct PDS for dead time
#
# Revnivtsev, Gilfanov & Churazov, 2000, A&A, 363, 1013
#
# DeadPDS(\@pds, $nonVLE_cnt, $VLE_cnt, $dtnb, $detnum);
#
# @pds - PDS array containing ($f, $df, $p, $dp) bins.
#        This is a standard PDS created by powspec with
#        normalization = -2.
#
# The next two parameters can be calculated from a lightcurve created
# from Standard1 files
#
# $nonVLE_cnt - average count rate from the following columns:
#   XeCntPcu0
#   XeCntPcu1
#   XeCntPcu2
#   XeCntPcu3
#   XeCntPcu4
#   RemainingCnt
#   VpCnt
#
# $VLE_cnt - average count rate from VLEcnt column
#
# $dtnb - time binning (in seconds)
#
# $detnum - number of PCU detectors used
#
{
  my $pds = $_[0];
  my $nonVLE_cnt = $_[1];
  my $VLE_cnt = $_[2];
  my $binsz = $_[3];
  my $detnum = $_[4];

  my $tau_d = 8.7e-6;
  my $tau_vle = 76e-6;
  my ($bin, $f, $df, $p, $dp, $r, $rp, $mu_VLE, $corrp);
  my ($sum, $k, $i, $omega, $otd, $X);
  my ($p_dt, $p_VLE);

  $mu_VLE = 1 / (1 - $VLE_cnt * $tau_vle);

  $r = $nonVLE_cnt * $mu_VLE;
  $rp = $nonVLE_cnt / (1 - $nonVLE_cnt * $tau_d);

  $i = 0;
  foreach $bin (@$pds)
  {
    ($f, $df, $p, $dp) = @$bin;
    $sum = 0;
    for($k = -10; $k <= 10; $k++)
    {
      $omega = 2 * $PI * ($f + $k / $binsz);
      $otd = $omega * $tau_d;
      $X = $rp*$rp * (1 - cos($otd)) + $rp * $omega * sin($otd);
      $X /= $rp*$rp * (1 - cos($otd))**2 + ($rp * sin($otd) + $omega)**2;
      $sum += $X / ($PI * ($f * $binsz + $k))**2;
    }

    $p_dt = -4 * (1/$r) * (sin(2 * $PI * $f * $binsz / 2))**2 * $sum;
    $p_VLE = 2 * $VLE_cnt * $tau_vle * $tau_vle * (sin($PI * $tau_vle * $f) / ($PI * $tau_vle *$f))**2;

    $corrp = 1/$detnum * ($p_dt * $mu_VLE + $p_VLE);
    $pds->[$i][2] = $p - $corrp;
    $i++;
  }
}

################################################

sub RmsPds
#
# Integrate power density spectrum and find total rms
#
#   ($rms, $err) = RmsPds(\@pds);
#
# INPUT
#   $pds - array with the PDS (freq, df, p, dp)
#
# OUTPUT
#   $rms - fractional rms integrated over the whole PDS
#   $err - its error
# 
{
  my $pds = $_[0];
  my ($sumrms, $sumerr, $rms, $err, $bin, $f, $df, $p, $dp);

  $sumrms = 0;
  $sumerr = 0;
  foreach $bin (@$pds)
  {
    ($f, $df, $p, $dp) = @$bin;
    $sumrms += $p * 2 * $df;
    $sumerr += ($dp * 2 * $df)**2;
  }
  if($sumrms > 0)
  {
    $rms = sqrt($sumrms);
    $err = sqrt($sumerr) / (2 * $rms);
  }
  else
  {
    $rms = 0;
    $err = 0;
  }
  return($rms, $err);
}

#######################################

sub RmsPdsBkg
#
# Integrate power density spectrum and find total rms
# correcting (rms/mean)^2 for background. This should be
# used with NOT background subtracted lightcurves.
#
#   ($rms, $err) = RmsPds(\@pds, $bkg);
#
# INPUT
#   $pds - array with the PDS (freq, df, p, dp)
#   $bkg - average background count rate
#
# OUTPUT
#   $rms - fractional rms integrated over the whole PDS
#   $err - its error
# 
{
  my $pds = $_[0];
  my $bkg = $_[1];
  my ($sumrms, $sumerr, $rms, $err, $bin, $f, $df, $p, $dp);

  $sumrms = 0;
  $sumerr = 0;
  foreach $bin (@$pds)
  {
    ($f, $df, $p, $dp) = @$bin;
    $sumrms += $p * 2 * $df;
    $sumerr += ($dp * 2 * $df)**2;
  }
  if($sumrms > 0)
  {
    $rms = sqrt($sumrms);
    $err = sqrt($sumerr) / (2 * $rms);
  }
  else
  {
    $rms = 0;
    $err = 0;
  }
  return($rms, $err);
}

################################################

sub RmsPds2
#
# Integrate power density spectrum and find total rms
#
#   ($rms2, $err, $min, $max) = RmsPds2(\@pds, $fmin, $fmax);
#
# INPUT
#   $pds - array with the PDS (freq, df, p, dp)
#
# OUTPUT
#   $rms - fractional rms integrated from $fmin to $fmax
#   $err - its error
# 
{
  my ($pds, $fmin, $fmax) = @_;
  my ($sumrms, $sumerr, $rms, $err, $bin, $f, $df, $p, $dp, $delta);

  my $fmin0 = $pds->[0][0] - $pds->[0][1];
  my $fmax0 = $pds->[@$pds - 1][0] + $pds->[@$pds - 1][1];

  $fmin = $fmin0 if($fmin < $fmin0);
  $fmax = $fmax0 if($fmax > $fmax0 || $fmax == 0);  
 
  return(0, 0, $fmin, $fmax) if($fmin >= $fmax);
    
  $sumrms = 0;
  $sumerr = 0;
  foreach $bin (@$pds)
  {
    ($f, $df, $p, $dp) = @$bin;
    next if($f + $df < $fmin || $f - $df > $fmax);

    $delta = Min($f + $df, $fmax) - Max($f - $df, $fmin);
    $sumrms += $p * $delta;
    $sumerr += ($dp * $delta)**2;
  }
  if($sumrms > 0)
  {
    $rms = sqrt($sumrms);
    $err = sqrt($sumerr) / (2 * $rms);
  }
  else
  {
#    $rms = 0;
#    $err = 0;
    $rms = -sqrt(abs($sumrms));
    $err = abs(sqrt(abs($sumerr)) / (2 * $rms));
  }
  return($rms, $err, $fmin, $fmax);
}

sub Max
{
  return($_[0] > $_[1] ? $_[0] : $_[1]);
}

sub Min
{  
  return($_[0] > $_[1] ? $_[1] : $_[0]);
}

#############################################################################

sub StatLc
#
# Computes intrinsic rms following Vaughan et al. 2003, MNRAS, 345, 1271
#
#  ($avg, $sig, $sig_err, $fvar, $fvar_err) = StatLc($filename);
#
#  $filename - light curve
#
#  $sig - standard deviation
#  $fvar - intrincic standard deviation / measn
#
{
  my $file_name = $_[0];
  
  my ($avg, $sig, $sig_err, $sig2, $sig2_err, $sig2_xs, $sig2_nxs, $err_sig2_nxs, $fvar, $err_fvar);
  my ($bin, $time, $cnt, $err);
  my ($n, $chi, $s_cnt, $s_chi2, $s_err2);
  my @lcurve = ();


  @lcurve = FitsDump("${file_name}+1", "TIME RATE ERROR");
  my @key = grep /=/, `fkeyprint infile=${file_name}+1 keynam=TIMEDEL outfile=STDOUT`;
  $key[0] =~ /=\s+([\d\.E\+]+)\s+/;
  my $binsize = $1;


  $s_cnt = 0;
  $n = 0;
  foreach $bin (@lcurve)
  {
    ($time, $cnt, $err) = @$bin;
    next if $cnt eq "INDEF";
    $s_cnt += $cnt;
    $n++;
  }
  $avg = $s_cnt / $n;

  $s_chi2 = 0;
  $s_err2 = 0;
  foreach $bin (@lcurve)
  {
    ($time, $cnt, $err) = @$bin;
    next if $cnt eq "INDEF";
    if($err * $binsize < 0.1) {$err = 1 / $binsize;}  # artificial error
    $chi = $cnt - $avg;
    $s_chi2 += $chi * $chi;
    $s_err2 += $err * $err;
  }

  $sig2 = $s_chi2 / ($n - 1);
  $sig = sqrt($sig2);
  $sig2_err = $s_err2 / $n;
  $sig_err = sqrt($sig2_err);

  $sig2_xs = $sig2 - $sig2_err;
  $sig2_nxs = $sig2_xs / $avg**2;  
  $fvar = $sig2_xs > 0 ? sqrt($sig2_xs) / $avg : 0;

  $err_sig2_nxs = sqrt(2/$n * $sig2_err**2 / $avg**4 + $sig2_err/$n * 4*$fvar**2/$avg**2);
  $err_fvar = $fvar > 0 ? 1 / (2 * $fvar) * $err_sig2_nxs : sqrt($err_sig2_nxs);
  
  return($avg, $sig, $sig_err, $fvar, $err_fvar);
}




1;
