package Xray::RXTE;
require Exporter;

use strict;
use warnings;

use File::Copy;
use Xray::FITS;
use Xray::Tools;

our $VERSION = 2.01;

our @ISA = ("Exporter");
our @EXPORT = qw(
 AnalyzeMode
 CalCol
 ChanEn
 DataMode
 Days
 FilesIntervals
 GetFileNums
 GetModes
 Inidir
 MkGTI
 MkFilt
 ObsFileList
 PCAcols
 PCUpds
 PCUon
 PhaDate
 ReadIntervals
 ReadStartStop
 RunPcaBackEst
 WhichPDSModes
$AUX_DIR
$BARY_DIR
$BKG_DIR
$BKH_DIR
$SRC_DIR
$DAT_DIR
$EV_DIR
$EVF_DIR
$FLT_DIR
$FRQ_DIR
$GX_DIR
$INT_DIR
$LC_DIR
$POW_DIR
$RXTE_data_path
@hxtcoeffs
@hxtresps
$Mbright
$Mfaint
$obsid_regexp
$standard_pca_expr
$saa_file
);

our $AUX_DIR = './_aux';
our $BARY_DIR = './_bary';
our $BKG_DIR = './_bkg';
our $BKH_DIR = './_bkh';
our $SRC_DIR = './_src';
our $DAT_DIR = './_data';
our $EV_DIR = './_ev';
our $EVF_DIR = './_evf';
our $FLT_DIR = './_flt';
our $FRQ_DIR = './_freq';
our $GX_DIR = './_gx';
our $INT_DIR = './_int';
our $LC_DIR = './_lc';
our $POW_DIR = './_pow';

our $RXTE_data_path = $INC{'Xray/RXTE.pm'};
$RXTE_data_path =~ s/RXTE\.pm/dat/;

our $Mbright = 'pca_bkgd_cmbrightvle_eMv20051128.mdl';
our $Mfaint = 'pca_bkgd_cmfaintl7_eMv20051128.mdl';
our $obsid_regexp = '\d\d\d\d\d\-\d\d-\d\d-\d\d\d?[SGR]?';
our $standard_pca_expr = 'ELV > 10.0 && OFFSET < 0.02 && (TIME_SINCE_SAA < 0 || TIME_SINCE_SAA > 30)';
our $saa_file = "${RXTE_data_path}/pca_saa_history";
our @hxtcoeffs = qw(hxtdead_200002_pwa.fits hxtdead_200002_pwb.fits);
our @hxtresps = qw(HXT_A.rsp HXT_B.rsp);

####################################################

sub AnalyzeMode
#
# Analyzes mode descriptor
#
# ($id, $binsize, $chnum, $chmin, $chmax) = AnalyzeMode($mode);
# 
# $mode - mode descriptor, e.g. B_2ms_8B_0_35_Q, can be read using DataMode sub.
#
# $id - configuration type: B, E, SB, GX1, GX2 or X, X meaning not supported
# $binsize - bin size in seconds
# $chnum - number of available energy channels
# $chmin - first channel
# $chmax - last channel
#
{
  my $mode = shift;

  my ($read);
  my ($i, $id, $binsize, $chnum, $chmin, $chmax);

  my %time;
  my @timearr = ("1s", "500ms", "250ms", "125ms", "62ms", "32ms", "16ms", "8ms", "4ms", "2ms", "1ms", "500us", "250us", "125us", "62us", "31us", "16us", "8us", "4us", "2us", "1us");
  $i = 0;
  for (@timearr)
  {
    $time{$_} = 1 / (2**$i);
    $i++;
  }
  
  if($mode =~ /^E/)
  {
    ($id, $binsize, $chnum, $chmin, $read) = split("_", $mode);
    $chmax = 249;
  }
  elsif($mode =~ /^B/)
    {($id, $binsize, $chnum, $chmin, $chmax, $read) = split("_", $mode);}
  elsif($mode =~ /^SB/)
  {
    ($id, $binsize, $chmin, $chmax, $read) = split("_", $mode);
    $chnum = "1";
  }
  elsif($mode =~ /GoodXenon/)
  {
    ($id, $read) = split("_", $mode);
    $id = ($id =~ /1/) ? 'GX1' : 'GX2';
    $binsize = 1 / (2**20);
    $chmin = 0;
    $chmax = 255;
    $chnum = 256;
    return($id, $binsize, $chnum, $chmin, $chmax);
  }
  else
    {return("X", 0, 0, 0, 0);}

  $binsize =~ tr/A-Z/a-z/;  # convert to lowercase
  if($binsize =~ /\ds/)
    {$binsize =~ s/\D//;}
  else
    {$binsize = $time{$binsize};}
  $chnum =~ s/\D//;
  return($id, $binsize, $chnum, $chmin, $chmax);
}
###################################################################

sub CalCol
#
# ($sc, $hc) = CalCol(\@spectrum, $time);
#
# Colour corrections valid only for one particular date in epoch 4!!!
#
{
  my $spectrum = $_[0];
  my $time = $_[1];

  my @E = (63504002, 70050782, 72227102, 164741884, 200793604, 241573804);
  my @ebound = (3.0, 4.0, 6.4, 9.7, 16.0);

  my ($epoch, $begin_epoch, $end_epoch, $rel_time);
  my ($chen, $cherr, $cnt, $cnterr);
  my ($e1, $e2, $min_energy, $max_energy, $frac);
  my ($i, $band);
  my @band_cnt;

  $i = 1;
  while($time > $E[$i])
    {$i++;}

  if($i > $E[5]) {die;}
  $epoch = $i;
  $begin_epoch = $E[$i - 1];
  $end_epoch = $E[$i];
  $rel_time = ($time - $begin_epoch) / ($end_epoch - $begin_epoch);

  @band_cnt = (0, 0, 0, 0);
  for($band = 0; $band <= 3; $band++)
   {
      $e1 = $ebound[$band];
      $e2 = $ebound[$band + 1];
      foreach my $s (@$spectrum)
      {
        ($chen, $cherr, $cnt, $cnterr) = @$s;
        $min_energy = $chen - $cherr;
        $max_energy = $chen + $cherr;
        if($max_energy > $e1 && $min_energy < $e2)   # the channel is in the band
          {
            $frac = 1;
            if($min_energy < $e1)  # interpolate at lower end
              {$frac = ($max_energy - $e1) / ($max_energy - $min_energy);}
            if($max_energy > $e2)  # interpolate at higher end
              {$frac = ($e2 - $min_energy) / ($max_energy - $min_energy);}

            $cnt = $frac * $cnt;  # scaling
            $band_cnt[$band] += $cnt;
         }
      }
    }

  my($sc, $hc);

  $sc = $band_cnt[1] / $band_cnt[0];
  $hc = $band_cnt[3] / $band_cnt[2];

  my @sfac = (0, 0, 0.045, 0, 0.01);
  my @hfac = (0, 0, -0.038, 0, -0.035);
  my @sshift = (-0.03, 0.145, -0.071, 0.005, 0.135);
  my @hshift = (0.037, 0.042, 0.032, 0.005, 0.005);

  $i = $epoch - 1;
  $sc = $sc * (1 + $sfac[$i] * $rel_time) + $sshift[$i];
  $hc = $hc * (1 + $hfac[$i] * $rel_time) + $hshift[$i];

  return($sc, $hc);
}

###########################################################################

sub ChanEn
#
# @engy = ChanEn($afile, $epoch);
#
# Returns epoch and (roughly - not corrected for gain drift)
# energy boundaries for a given file, $afile.
#
# $engy[$ch] is the upper energy boundary for channel $ch
#
{
  my $filename = $_[0];

  my @e1 = (1.51,1.51,1.51,1.51,1.51,1.76,2.02,2.27,2.53,2.78,3.04,3.30,3.55,3.81,4.07,4.32,4.58,4.84,5.10,5.35,5.61,5.87,6.13,6.39,6.65,6.91,7.17,7.43,7.69,7.95,8.21,8.47,8.73,8.99,9.25,9.52,9.78,10.04,10.30,10.57,10.83,11.09,11.36,11.62,11.89,12.15,12.42,12.68,12.95,13.21,13.48,13.74,14.01,14.28,14.81,14.81,15.35,15.35,15.88,15.88,16.42,16.42,16.96,16.96,17.50,17.50,18.04,18.04,18.58,18.58,19.12,19.12,19.67,19.67,20.21,20.21,20.76,20.76,21.30,21.30,21.85,21.85,22.40,22.40,22.95,22.95,23.50,23.50,24.06,24.06,24.61,24.61,25.17,25.17,25.72,25.72,26.28,26.28,26.84,26.84,27.40,27.40,27.96,27.96,28.52,28.52,29.09,29.09,29.65,29.65,30.22,30.22,30.79,30.79,31.36,31.36,31.93,31.93,32.50,32.50,33.07,33.07,33.64,33.64,34.22,34.22,34.80,34.80,35.38,35.38,35.95,35.95,36.54,36.54,37.12,37.12,37.99,37.99,37.99,38.87,38.87,38.87,39.75,39.75,39.75,40.64,40.64,40.64,41.53,41.53,41.53,42.42,42.42,42.42,43.32,43.32,43.32,44.21,44.21,44.21,45.12,45.12,45.12,46.02,46.02,46.02,46.93,46.93,46.93,47.84,47.84,47.84,48.76,48.76,48.76,49.68,49.68,49.68,50.60,50.60,50.60,51.53,51.53,51.53,52.46,52.46,52.46,53.39,53.39,53.39,54.33,54.33,54.33,55.27,55.27,55.27,56.22,56.22,56.22,57.17,57.17,57.17,58.12,58.12,58.12,59.08,59.08,59.08,60.04,60.04,60.04,61.00,61.00,61.00,61.97,61.97,61.97,62.95,62.95,62.95,63.93,63.93,63.93,64.91,64.91,64.91,65.90,65.90,65.90,66.89,66.89,66.89,67.89,67.89,67.89,68.89,68.89,68.89,70.23,70.23,70.23,70.23,71.58,71.58,71.58,71.58,72.94,72.94,72.94,72.94,74.99,74.99,74.99,74.99,74.99,74.99);

  my @e2 = (1.61,1.61,1.61,1.61,1.61,1.91,2.22,2.52,2.83,3.13,3.44,3.75,4.05,4.36,4.67,4.98,5.28,5.59,5.90,6.21,6.52,6.83,7.14,7.45,7.76,8.07,8.38,8.69,9.00,9.31,9.63,9.94,10.25,10.56,10.88,11.19,11.50,11.82,12.13,12.45,12.76,13.08,13.39,13.71,14.03,14.34,14.66,14.98,15.29,15.61,15.93,16.25,16.57,16.89,17.52,17.52,18.16,18.16,18.81,18.81,19.45,19.45,20.09,20.09,20.74,20.74,21.39,21.39,22.03,22.03,22.68,22.68,23.33,23.33,23.99,23.99,24.64,24.64,25.29,25.29,25.95,25.95,26.61,26.61,27.27,27.27,27.93,27.93,28.59,28.59,29.25,29.25,29.91,29.91,30.58,30.58,31.25,31.25,31.92,31.92,32.59,32.59,33.26,33.26,33.93,33.93,34.61,34.61,35.28,35.28,35.96,35.96,36.64,36.64,37.32,37.32,38.00,38.00,38.68,38.68,39.37,39.37,40.06,40.06,40.74,40.74,41.43,41.43,42.13,42.13,42.82,42.82,43.51,43.51,44.21,44.21,45.26,45.26,45.26,46.31,46.31,46.31,47.36,47.36,47.36,48.42,48.42,48.42,49.49,49.49,49.49,50.55,50.55,50.55,51.62,51.62,51.62,52.70,52.70,52.70,53.78,53.78,53.78,54.86,54.86,54.86,55.95,55.95,55.95,57.04,57.04,57.04,58.13,58.13,58.13,59.23,59.23,59.23,60.33,60.33,60.33,61.44,61.44,61.44,62.55,62.55,62.55,63.67,63.67,63.67,64.79,64.79,64.79,65.92,65.92,65.92,67.05,67.05,67.05,68.18,68.18,68.18,69.32,69.32,69.32,70.47,70.47,70.47,71.62,71.62,71.62,72.77,72.77,72.77,73.93,73.93,73.93,75.10,75.10,75.10,76.27,76.27,76.27,77.44,77.44,77.44,78.62,78.62,78.62,79.81,79.81,79.81,81.00,81.00,81.00,82.20,82.20,82.20,83.80,83.80,83.80,83.80,85.42,85.42,85.42,85.42,87.04,87.04,87.04,87.04,89.50,89.50,89.50,89.50,89.50,89.50);


  my @e3 = (1.94,1.94,1.94,1.94,1.94,2.29,2.64,2.99,3.35,3.70,4.05,4.41,4.76,5.12,5.47,5.82,6.18,6.54,6.89,7.25,7.60,7.96,8.32,8.68,9.03,9.39,9.75,10.11,10.47,10.83,11.19,11.55,11.91,12.27,12.63,12.99,13.36,13.72,14.08,14.44,14.81,15.17,15.54,15.90,16.26,16.63,17.00,17.36,17.73,18.09,18.46,18.83,19.20,19.56,20.30,20.30,21.04,21.04,21.78,21.78,22.52,22.52,23.26,23.26,24.01,24.01,24.75,24.75,25.50,25.50,26.25,26.25,27.00,27.00,27.75,27.75,28.50,28.50,29.26,29.26,30.02,30.02,30.77,30.77,31.53,31.53,32.29,32.29,33.06,33.06,33.82,33.82,34.59,34.59,35.35,35.35,36.12,36.12,36.89,36.89,37.67,37.67,38.44,38.44,39.22,39.22,39.99,39.99,40.77,40.77,41.55,41.55,42.34,42.34,43.12,43.12,43.91,43.91,44.70,44.70,45.49,45.49,46.28,46.28,47.07,47.07,47.87,47.87,48.66,48.66,49.46,49.46,50.26,50.26,51.06,51.06,52.27,52.27,52.27,53.48,53.48,53.48,54.70,54.70,54.70,55.92,55.92,55.92,57.14,57.14,57.14,58.37,58.37,58.37,59.60,59.60,59.60,60.84,60.84,60.84,62.08,62.08,62.08,63.33,63.33,63.33,64.58,64.58,64.58,65.84,65.84,65.84,67.10,67.10,67.10,68.37,68.37,68.37,69.64,69.64,69.64,70.91,70.91,70.91,72.19,72.19,72.19,73.48,73.48,73.48,74.77,74.77,74.77,76.07,76.07,76.07,77.37,77.37,77.37,78.68,78.68,78.68,79.99,79.99,79.99,81.30,81.30,81.30,82.63,82.63,82.63,83.96,83.96,83.96,85.29,85.29,85.29,86.63,86.63,86.63,87.98,87.98,87.98,89.33,89.33,89.33,90.69,90.69,90.69,92.05,92.05,92.05,93.42,93.42,93.42,94.80,94.80,94.80,96.64,96.64,96.64,96.64,98.50,98.50,98.50,98.50,100.37,100.37,100.37,100.37,103.19,103.19,103.19,103.19,103.19,103.19);

  my @e4 = (2.13,2.13,2.13,2.13,2.13,2.54,2.96,3.37,3.79,4.21,4.63,5.04,5.46,5.88,6.30,6.72,7.14,7.56,7.98,8.40,8.82,9.24,9.67,10.09,10.51,10.93,11.36,11.78,12.21,12.63,13.06,13.48,13.91,14.34,14.76,15.19,15.62,16.05,16.47,16.90,17.33,17.76,18.19,18.62,19.05,19.49,19.92,20.35,20.78,21.22,21.65,22.08,22.52,22.95,23.82,23.82,24.70,24.70,25.57,25.57,26.45,26.45,27.33,27.33,28.21,28.21,29.09,29.09,29.97,29.97,30.86,30.86,31.74,31.74,32.63,32.63,33.52,33.52,34.42,34.42,35.31,35.31,36.21,36.21,37.10,37.10,38.00,38.00,38.91,38.91,39.81,39.81,40.72,40.72,41.62,41.62,42.53,42.53,43.44,43.44,44.36,44.36,45.27,45.27,46.19,46.19,47.11,47.11,48.03,48.03,48.96,48.96,49.88,49.88,50.81,50.81,51.74,51.74,52.67,52.67,53.61,53.61,54.54,54.54,55.48,55.48,56.42,56.42,57.37,57.37,58.31,58.31,59.26,59.26,60.21,60.21,61.64,61.64,61.64,63.07,63.07,63.07,64.51,64.51,64.51,65.95,65.95,65.95,67.40,67.40,67.40,68.86,68.86,68.86,70.32,70.32,70.32,71.78,71.78,71.78,73.25,73.25,73.25,74.73,74.73,74.73,76.21,76.21,76.21,77.70,77.70,77.70,79.19,79.19,79.19,80.69,80.69,80.69,82.20,82.20,82.20,83.71,83.71,83.71,85.23,85.23,85.23,86.75,86.75,86.75,88.28,88.28,88.28,89.81,89.81,89.81,91.36,91.36,91.36,92.91,92.91,92.91,94.46,94.46,94.46,96.02,96.02,96.02,97.59,97.59,97.59,99.17,99.17,99.17,100.75,100.75,100.75,102.34,102.34,102.34,103.93,103.93,103.93,105.54,105.54,105.54,107.15,107.15,107.15,108.76,108.76,108.76,110.39,110.39,110.39,112.02,112.02,112.02,114.21,114.21,114.21,114.21,116.41,116.41,116.41,116.41,118.63,118.63,118.63,118.63,121.98,121.98,121.98,121.98,121.98,121.98);

  my @e5 = (2.06,2.06,2.06,2.06,2.06,2.47,2.87,3.28,3.68,4.09,4.49,4.90,5.31,5.71,6.12,6.53,6.94,7.35,7.76,8.17,8.57,8.98,9.40,9.81,10.22,10.63,11.04,11.45,11.87,12.28,12.69,13.11,13.52,13.93,14.35,14.76,15.18,15.60,16.01,16.43,16.85,17.26,17.68,18.10,18.52,18.94,19.36,19.78,20.20,20.62,21.04,21.46,21.88,22.30,23.15,23.15,24.00,24.00,24.85,24.85,25.70,25.70,26.55,26.55,27.40,27.40,28.26,28.26,29.12,29.12,29.97,29.97,30.83,30.83,31.70,31.70,32.56,32.56,33.43,33.43,34.29,34.29,35.16,35.16,36.03,36.03,36.91,36.91,37.78,37.78,38.66,38.66,39.53,39.53,40.41,40.41,41.30,41.30,42.18,42.18,43.06,43.06,43.95,43.95,44.84,44.84,45.73,45.73,46.62,46.62,47.52,47.52,48.41,48.41,49.31,49.31,50.21,50.21,51.12,51.12,52.02,52.02,52.93,52.93,53.83,53.83,54.74,54.74,55.66,55.66,56.57,56.57,57.49,57.49,58.40,58.40,59.78,59.78,59.78,61.17,61.17,61.17,62.56,62.56,62.56,63.96,63.96,63.96,65.36,65.36,65.36,66.76,66.76,66.76,68.17,68.17,68.17,69.59,69.59,69.59,71.01,71.01,71.01,72.43,72.43,72.43,73.86,73.86,73.86,75.30,75.30,75.30,76.74,76.74,76.74,78.18,78.18,78.18,79.63,79.63,79.63,81.09,81.09,81.09,82.55,82.55,82.55,84.02,84.02,84.02,85.49,85.49,85.49,86.97,86.97,86.97,88.46,88.46,88.46,89.95,89.95,89.95,91.45,91.45,91.45,92.95,92.95,92.95,94.46,94.46,94.46,95.97,95.97,95.97,97.49,97.49,97.49,99.02,99.02,99.02,100.55,100.55,100.55,102.09,102.09,102.09,103.64,103.64,103.64,105.19,105.19,105.19,106.75,106.75,106.75,108.32,108.32,108.32,110.42,110.42,110.42,110.42,112.53,112.53,112.53,112.53,114.65,114.65,114.65,114.65,117.86,117.86,117.86,117.86,117.86,117.86);

  my @e = ([@e1], [@e2], [@e3], [@e4], [@e5]);

  my @epoch_boundaries = (70050782, 72227102, 164741884, 200793604, 1e16);
  my ($key, $tstart, $epoch);


  $key = `fkeyprint infile=$filename keynam=TSTART outfile="STDOUT" | grep =`;
  $key =~ /=\s+(.*) \//;
  $tstart = $1;
  $epoch = 0;
  while($epoch_boundaries[$epoch] < $tstart)
    {$epoch++;}
  $epoch++;

  $_[1] = $epoch;
  return(@{$e[$epoch - 1]});
}

#######################################################

sub DataMode
#
# Reads and returns DATAMODE key from the fits file
#
#   $mode = DataMode($file_name)
#
# INPUT
#
#   $file_name - name of the data file
#
#
# OUTPUT
#
#   $mode - mode string
#
{
  my $file_name = $_[0];

  my ($mode);

  $mode = `fkeyprint infile=$file_name keynam=DATAMODE outfile="STDOUT" | grep =`;
  chomp $mode;
  $mode =~ s/(.*)=//;
  $mode =~ s/\/(.*)//;
  $mode =~ s/ //;
  $mode =~ s/\'//g;
  return($mode);
}


################################################################

sub Days
#
# Divide the list of files into days. 
#
#   Days(\%timefile, $fileday, $day);
#
# INPUT:
#
#   timefile - hash table of files (TSTART -> filename), 
#              created by StartTime
#
# OUTPUT:
#
#   fileday - hash table of days: (filename -> day)
#   day - list of days, sorted
#
{
  my $timefile = $_[0];

  my %fileday = ();
  my @day = ();
  my $old_day = -1;
  my ($tstart, $d, $di);

  foreach $tstart (sort keys %$timefile)
  {
    $d = $tstart / 86400.0;
    $di = int($d);
    if($d - $di < 0.34) {$di--;}
    $fileday{$timefile->{$tstart}} = $di;
    if($di != $old_day)
    {
      push(@day, $di);
      $old_day = $di;
    }
  }
  $_[1] = \%fileday; 
  $_[2] = \@day;
}    


################################################

sub FilesIntervals
#
# @fileint = FilesIntervals(\@filetime, \@intervals);
#
# The list has form:
#   filename, ((begin, end), (begin, end), ...)
#   filename, ((begin, end), (begin, end), ...)
# ....
{
  my $filetime = $_[0];
  my $intervals = $_[1];
  my @fileint = ();
  my ($f, $t, $name, $tstart, $tstop, $begin, $end, $cnt, $totcnt);
  my @ints;

  $totcnt = 0;
  foreach $f (@$filetime)
  {
    ($name, $tstart, $tstop) = @$f;
    @ints = ();
    $cnt = 0;
    foreach $t (@$intervals)
    {
      ($begin, $end) = @$t;
      if($begin > $tstart && $end < $tstop)
      {
        push @ints, [($begin, $end)];
        $cnt++;
        $totcnt++;
      }
    }
    if($cnt != 0)
      { push @fileint, [($name, [@ints])]; }
    $name =~ /(.*)(pca|hexte)\/(.*)/;
#    print "$3 -- $cnt intervals found.\n";
  }
#  print "\nTotal number of $totcnt intervals found.\n";
  return @fileint;
}
 
################################################################

sub GetFileNums
#
# Create list of filenums, obsids and a cross-reference hash
#
#     GetFileNums(\@file_list, $filenums, $obsids, $namenum, $numid);
#
# INPUT:
# 
#   file_list - list of pathnames (e.g. created by ReadFileList)
#
# OUTPUT:
#
#   filenums - list of filenums
#   obsids - list of obsids
#   namenum - hash (filename -> filenum)
#   numid - hash (filenum -> obsid)
#
# EXAMPLE:
#
# From the list of pathnames:
#
#   /data/P30411/30411-01-03-00/pca/FS4a_811aed0-811bf00
#   /data/P30411/30411-01-03-00/pca/FS4a_811c6b0-811d03e
#   /data/P30411/30411-01-04-00/pca/FS4a_811f63d-8120730
#
# a list of the following filenums will be created:
#
#   30411-01-03-00a
#   30411-01-03-00b
#   30411-01-04-00
#
# and obsids:
#
#   30411-01-03-00
#   30411-01-04-00
#
{
  my $file_list = $_[0];  # this is a reference to an array!

  my @file_nums = ();
  my @obsids = ();
  my %numid = ();
  my %namenum = ();
  my $old_file_name = "0";
  my $old_num = "0";
  my $old_obsid = "0";
  my $diff_obsid = "0";
  my $letter = 'a';

  my ($i, $obsid, $num, $file_name, $file_num);

  foreach $file_name (@$file_list)
  {
    if ($file_name =~ /(.*)(P.....)\/(.{14}[0-9GRS]?)\/(pca|hexte)/)
    {
      $obsid = $3;
      if ($obsid ne $diff_obsid)
      {
        push(@obsids, $obsid);
        $diff_obsid = $obsid;
      }
#      $obsid =~ /(.{9})(.{5})/;
#      $num = $2;
      $num = $obsid;
      if ($old_num ne "0")
      {
        if($old_num eq $num)
        {
          $file_num = $old_num . $letter;
          ++$letter;
        }
        else   
        {
          $file_num =  $old_num . (($letter eq 'a') ? '' : $letter);
          $letter = 'a';
        }
        push(@file_nums, $file_num);
        $numid{$file_num} = $old_obsid;
        $namenum{$old_file_name} = $file_num;
      }
      $old_num = $num;
      $old_obsid = $obsid;
      $old_file_name = $file_name;
    }
    else
    {
      die("\n*** Incorrect pathname in the data file:\n$file_name\n\n");
    }
  }
  $file_num =  $old_num . (($letter eq 'a') ? '' : $letter);  
  push(@file_nums, $file_num);
  $numid{$file_num} = $obsid;
  $namenum{$old_file_name} = $file_num;
  $_[1] = \@file_nums;
  $_[2] = \@obsids;
  $_[3] = \%namenum;
  $_[4] = \%numid;
}
################################################

sub GetModes
#
# Gets a list of modes from config.lst file for a given obsid.
#
#  @mlist = GetModes($obsid);
#
# @mlist contains mode names, like E_16us_64M_24_1s
#
{
  my $obsid = shift;

  my @mlist = ();
  my $line;

  if(!-e "config.lst")
    {die "config.lst not found.\n";}

  $line = `grep \"${obsid} \" config.lst`;
  if($line eq "")
    {return @mlist;}

  @mlist = split(" ", $line);
  shift @mlist;
  return(@mlist);
}

################################################

sub Inidir
#
# Inidir($dir1, $dir2, ..., $dirn);
#
{
  foreach my $dir (@_)
  {
    if (!(-d $dir)) 
    {
      print "Creating $dir.\n";
      mkdir("$dir", 0755);
    }
  }
}

################################################################

sub MkFilt
#
# @flt = MkFilt($std2_list, $appids)
#
# bug fixed, 5-9-2001
{
  my $std2_list = $_[0];
  my $appids = $_[1];
  my $logfile = "/tmp/mkfilt$$.log";

  my (@std2_files, $path, $flt, $filenums, $namenum, $numid, $obsids, $args, $stat, $n);
  my (@log);

  Inidir($FLT_DIR);
  if(!(-e "$appids")) {die("\n*** File $appids does not exist!\n\n");}

  $| = 1;
  unlink $logfile;
  local *LOG;
  open(LOG, ">$logfile");

  @std2_files = ReadFileList($std2_list);
  GetFileNums(\@std2_files, $filenums, $obsids, $namenum, $numid);

  $n = 0;
  my @flt = ();
  foreach my $obsid (@$obsids)
  {
    while(!($std2_files[$n] =~ /$obsid/))
      { $n++; }
    $std2_files[$n] =~ /(.*P\d{5})/;
    $path = $1;
    $flt = "$FLT_DIR\/$obsid";
    push @flt, $flt;
    if(!(-e "$flt\.xfl"))
    {
      print "Creating filter file $flt\.xfl\n";
      $args = "xtefilt -a $appids -o $obsid -p $path -t 1. -f $flt -c y";
      @log = &call($args, $stat);
      print LOG @log;
      ToBe($stat, $logfile);
    }
    else {print "*** $flt\.xfl already exisits. Skipping XTEFILT...\n";}
  }
  close(LOG);
  return @flt;
}

################################################################

sub MkGTI
#
# MkGTI($std2_list, $gti_file, $expr_file);
#
{
  my ($std2_list, $gti, $expr_file) = @_;

  my @files = ReadFileList($std2_list);
  my %obsfiles = ObsFileList(\@files);
  my @obsids = sort keys(%obsfiles);
  
  my (@log, $args, $stat);
  my @garbage = ();
  
  local *LOG;
  my $log_file = "/tmp/mkgti$$.log";
  open(LOG, ">$log_file") or die "$!\n";
  
  local *FLT;
  open(FLT, 'Filter.lst') or die "\nFilter.lst not found. Did you run xtefiles?\n";
  my @flts = <FLT>;
    
  my @gtis = ();
  my $i = 0;
  my $nobs = @obsids;
  foreach my $obsid (@obsids)
  {
    my $gti_file = "/tmp/${obsid}_$$.gti";
    my @flt1 = grep /$obsid/, @flts;
    my $flt = $flt1[0];
    chomp $flt;
    push @gtis, $gti_file;
    push @garbage, $gti_file;

    #print "Working on obsid $obsid...\n";
    Percent(++$i/$nobs);
    $flt .= '.gz' if !-e $flt;
    die "\nFilter file $flt does not exist.\n" if !-e $flt;

    unlink $gti_file;
    $args = "maketime infile=$flt outfile=$gti_file expr=\@$expr_file name=NAME value=VALUE time=TIME compact=no";
    @log = call($args, $stat);
    print LOG @log;
    ToBe($stat, $log_file);
  }
  
  if(@obsids == 1)  # do not merge
  {
    print "\n";
    move $gtis[0], $gti;
    return;
  }

#
# Merge GTI filters into one file
#

  print "\nMerging GTI files\n";
  unlink $gti;

  local *GTI;
  my $groups = 0;
  my $n = 1;
  foreach my $gti_file (@gtis)
  {
    if($n % 400 == 1)
    {
      $groups++;
      if($groups > 1) {close(GTI);}
      my $grp_file = "/tmp/gti_${groups}_$$.lst";
      open(GTI, ">$grp_file") or die;
      push @garbage, $grp_file; 
    }
    $n++;
    print GTI "$gti_file\n";
  }
  close(GTI);

  local *GRP;
  my $grp_gti;
  my $final_list = "/tmp/gti_grp_$$.lst";
  push @garbage, $final_list;
  open(GRP, ">$final_list") or die;
  for $n (1 .. $groups)
  {
    $grp_gti = "/tmp/grp${n}_$$.gti";
    my $grp_list = "/tmp/gti_${n}_$$.lst";
    push @garbage, $grp_gti;
    push @garbage, $grp_list;
    unlink $grp_gti;
    $args = "mgtime \@$grp_list $grp_gti OR";
    @log = call($args, $stat);
    print LOG @log;
    ToBe($stat, $log_file);
    print GRP "$grp_gti\n";
  }
  close(GRP);

  if($groups > 1)
  {
    $args = "mgtime \@$final_list $gti OR";
    @log = call($args, $stat);
    print LOG @log;
    ToBe($stat, $log_file);
  }
  else
  {
    move $grp_gti, $gti;
  }
  for my $f (@garbage)
    {unlink $f;}
  
  close(LOG);
}

################################################################

sub ObsFileList
#
# Create a list of files for each obsid
#
#   %obsfiles = ObsFileList(\@files)
#
# INPUT:
#
#   @files - list of data file names (with paths)
#
# OUTPUT:
#
#   %obsfiles - hash obsid -> @list_of_files
#
# USAGE:
#
#   @obsids = sort keys(%obsfiles)  - list of obsids
#   foreach $f (@{$obsfiles{$obsid}) - loop all files for a given obsid
#
{
  my $files = $_[0];
  my %obsfiles = ();
  my ($f, $obsid);

  foreach $f (@$files)
  {
    chomp $f;
    $f =~ /($obsid_regexp)[\/\._a-z]/;
    $obsid = $1;
    push @{$obsfiles{$obsid}}, $f; 
  }
  return %obsfiles;
}

#######################################################

sub PCAcols
#
# Returns list of PCA columns for given detectors and layers
#
# @cols = PCAcols($detectors, $layers)
#
# $detectors - eg '0', '2,3', 'all'
# $layers - eg 'L1', 'LR1', 'LR1, LR2' or 'all'
#
{
  my ($detectors, $layers) = @_;
  
  $detectors =~ s/\s//g;
  $layers =~ s/\s//g;
  
  if($detectors eq 'all') {$detectors = '0,1,2,3,4';}
  if($layers eq 'all') {$layers = 'LR1,LR2,LR3';}

  $layers = uc $layers;
    
  my @dets = split /,/, $detectors;
  my @lays = split /,/, $layers;
  
  my @cols = ();
  foreach my $d (@dets)
  {
    foreach my $l (@lays)
    {
      if($l =~ /^([LR])(\d)/) 
        {push @cols, "X${2}${1}SpecPcu${d}";}
      elsif($l =~ /^LR(\d)/) 
        {
	  push @cols, "X${1}LSpecPcu${d}";
	  push @cols, "X${1}RSpecPcu${d}";
	}
    }
  }
  
  return @cols;
}

#######################################################

sub PCUpds
#
# Selcts best PCU configuration for PDS creation.
#
# @pcu = PCUpds($filterfile, $pertime)
#
# $filterfile - pathname of the filter file (must exist!)
# @pcu - best configuration, e.g. (0, 1, 1, 0, 1)
# $pertime - percentage of time for the best config.
#
{
  my $filterfile = $_[0];
  
  my @flt = FitsDump($filterfile, "TIME ELV OFFSET TIME_SINCE_SAA PCU0_ON PCU1_ON PCU2_ON PCU3_ON PCU4_ON");
  my @pcu_conf = (0) x 32;
  my $tot = @flt;
  FLT: for my $f (@flt)
  {
    my ($time, $elv, $offset, $saa, @pcuon) = @$f;
    next FLT if("@$f" =~ /INDEF/);
    next FLT if($elv <= 10 || $offset >= 0.02 || ($saa >= 0 && $saa <= 30));
    my $conf = 0;
    for my $i (0 .. 4)
      {$conf += $pcuon[$i] * 2**$i;}    # unique PCU configuration
    $pcu_conf[$conf]++;
  }
  my $max = -1;
  my $best_conf_i = -1;
  for my $i (1 .. 31)   # Do not check 0 - all detectors off
  {
    if($pcu_conf[$i] > $max)
    {
      $max = $pcu_conf[$i];
      $best_conf_i = $i;
    }
  }
  my @best_conf = ();
  for my $i (0 .. 4)
  {
    push @best_conf, $best_conf_i % 2;
    $best_conf_i = int($best_conf_i / 2);
  }
  my $perc = 100 * $max / $tot;
  $_[1] = $perc;
  return @best_conf;
}

#######################################################

sub PCUon
#
# Give percentage of live time of each PCU, for a given obsid
#
# @pcu_on = PCUon($obsid, $pcuact)
#
# @pcu_on contains five numbers between 0 and 1.
# @$pcuact - array with PCU activity versus time (filter file dump)
#
{
  my $obsid = $_[0];
  
  my ($file, $cnt, $p, $n);
  my @pcu_on = (0, 0, 0, 0, 0);
  my @pcuact;
  my @pcuact_clean = ();
  
  $file = "${FLT_DIR}/${obsid}.xfl";
  return(@pcu_on) if(!-e $file);
  @pcuact = FitsDump("$file", "time pcu0_on pcu1_on pcu2_on pcu3_on pcu4_on");
  $cnt = 0;
  foreach $p (@pcuact)
  {
    if($p->[1] !~ /INDEF/)
    {
      $cnt++;
      for $n (0 .. 4)
        {$pcu_on[$n]++ if(@$p[$n + 1] eq "1");}
      push @pcuact_clean, [@$p];
    }    
  }
  for $n (0 ..  4)
    {$pcu_on[$n] = int(100 * ${pcu_on[$n]}/$cnt);}
  $_[1] = \@pcuact_clean;
  return(@pcu_on);
}

#######################################################

sub PhaDate
{
# get the DATE-OBS value from the Spectrum extension of the PHA file
#  Parse its format (dd/mm/yy or yyyy-mm-ddThh:mm:ss) into yyyy-mm-dd

  my ($i,$x,$phafile,$phaobs,$day,$month,$year,$phadate,$stat);
  $phafile = $_[0];
  my @result=&call('fkeyprint infile="'.$phafile.'[1]" keynam="DATE-OBS" outfile="STDOUT" exact="no" clobber="yes"');

  $i = 0;
  while($result[$i] !~ /DATE-OBS=/)
    {$i++};
  $phaobs = $result[$i];

  $x = index($phaobs,"'");
  if (index($phaobs,"-",$x) == -1) {
      $day = substr($phaobs,$x+1,2);
      $month = substr($phaobs,$x+4,2);
      $year = substr($phaobs,$x+7,2);
      if ($year > 90) {$year = '19'.$year} 

      $phadate = $year.'-'.$month.'-'.$day;
       
  } else {
      $phadate = substr($phaobs,$x+1,10);    
  }
}

################################################

sub ReadIntervals
#
#  @intervals = ReadIntervals($filename);
#
{
  my $filename = $_[0];
  my @intervals = ();
  my @s = ();

  local *TINT;
  open(TINT, $filename) or die("Cannot open TINT file $filename.\n");
  while(<TINT>)
  {
    @s = split;
    push @intervals, [($s[0], $s[1])];
  }
  return @intervals;
}

################################################

sub ReadStartStop
#
# @time_list = ReadStartStop(\@file_list);
#
# Creates a list of (filename, TSTART, TSTOP)
#
{
  my $file_list = $_[0];
  my @time_list = ();
  my ($f, $tstart, $tstop);

  foreach $f (@$file_list)
  {
    system("fkeypar $f TSTART");
    $tstart = `pget fkeypar value`;
    chop $tstart;
    system("fkeypar $f TSTOP");
    $tstop = `pget fkeypar value`;
    chop $tstop;
    push @time_list, [($f, $tstart, $tstop)];
    $f =~ /(.*)(pca|hexte)\/(.*)/;
    print "$3  $tstart -- $tstop\n";
  }
  return @time_list;
}

################################################

sub RunPcaBackEst
#
# @back_list = RunPcaBackEst($flag, $std2_list, $bkg_models, $saa_file);
#
{
  my $flag = $_[0];
  my $STD2_LIST = $_[1];
  my $BKG_MODELS = $_[2];
  my $saa = $_[3];

  my @back_list = ();
  my (@std2_files, $filenums, $obsids, $namenum, $numid, $logfile, $stat);

  Inidir($BKG_DIR, $FLT_DIR);

  my($gaincorr, $fullspec);
  if($flag eq "ev")
  {
    $gaincorr = "yes";
    $fullspec = "yes";
  }
  else
  {    
    $gaincorr = "no";
    $fullspec = "no";
  }

#  if(!(-e $saa)) {die "\n*** $saa does not exist!\n\n";}
    
  @std2_files = ReadFileList($STD2_LIST);
  GetFileNums(\@std2_files, $filenums, $obsids, $namenum, $numid);

  $logfile = "/tmp/prepback$$.log";
  unlink $logfile;
  local *LOG;
  open(LOG, ">$logfile") or die;

  my $old_obsid = "0";
  foreach my $data_file (@std2_files)
  {
    my $file_num = $namenum->{$data_file};
    my $obsid = $numid->{$file_num};
    my $filt_file = "$FLT_DIR\/$obsid\.xfl";
    if(!-e $filt_file)
      {die("\n*** Filter file $filt_file does not exists! Run MKFILT.\n\n");}
    my $bkg = "$BKG_DIR\/$file_num\.bkg";
    push(@back_list, $bkg);
    if(!-e "$bkg")
    {
      print "Running PCABACKEST for $file_num\n";
      my $args = "pcabackest $data_file $bkg \@$BKG_MODELS $filt_file saahfile=$saa interval=16 layers=yes gaincorr=$gaincorr gcorrfile=caldb fullspec=$fullspec syserr=no timeslop=32";
      my @log = &call($args, my $stat);
      print LOG @log;
      ToBe($stat, $logfile);
    }
    else 
    {
      print "*** File $bkg already exists. Skipping PCABACKEST.\n";
    }
  }
  close(LOG);
  return @back_list;
}

################################################################

sub WhichPDSModes
#
# Creates a list of PCA modes (configurations) for creating PDS. Selection
# gives modes with best timing resolution, typically one Event and two
# Single Bit, or one Event and one Binned.
#
# @PDS = WhichPDSModes(\@mlist);
#
# @mlist - list of modes obtained by GetModes
#
# @PDS - list of selected modes, each element contains:
#        ($mode, $id, $binsize, $chnum, $chmin, $chmax)
#
{
  my $mlist = shift;

  my ($mode, $id, $binsize, $chnum, $chmin, $chmax);
  my (@arr, $m, $m_max);

  my $sbmin = 1000;
  my $emin = 1000;
  my $max = -1;

  my @E = ();
  my @B = ();
  my @SB = ();
  my @GX = ();
  my @PDS = ();

  foreach $mode (@$mlist)
  {
    next if $mode =~ /LLD|L1R1|_F$/;  # don't know what LLD is, but crashes
    @arr = AnalyzeMode($mode);
    if($arr[0] eq "E")
      {push @E, [$mode, @arr];}
    elsif($arr[0] eq "B")
      {push @B, [$mode, @arr];}
    elsif($arr[0] eq "SB") 
      {push @SB, [$mode, @arr];}
    elsif($arr[0] =~ /GX/)
      {push @GX, [$mode, @arr];}
  }

  if(@GX > 0)
  {    
    ($mode, $id, $binsize, $chnum, $chmin, $chmax) = @{$GX[0]};
    $mode =~ s/n1/n/;   # convert GoodXenon1_ns into GoodXenon_ns
    $id = 'GX';
    my @m = ($mode, $id, $binsize, $chnum, $chmin, $chmax);
    push @PDS, [@m];
    return @PDS;
  }
    
# look for an event mode with minimal channel

  foreach $m (@E)
  {
    ($mode, $id, $binsize, $chnum, $chmin, $chmax) = @$m;
    if($chmin < $emin)
    {
      $emin = $chmin;
      @PDS = [@$m];
    }
  }
  if($emin == 0) {return @PDS;}

# check minimum SB channel

  foreach $m (@SB)
  {
    ($mode, $id, $binsize, $chnum, $chmin, $chmax) = @$m;
    if($chmin < $sbmin) {$sbmin = $chmin;}
  }

# find all SB modes

  if($sbmin == 0 || @B == 0)
  {
    foreach $m (@SB) 
    {
      ($mode, $id, $binsize, $chnum, $chmin, $chmax) = @$m;
      push @PDS, [@$m] if $chmax < $emin;
    }
    return @PDS;
  }
  
# if there are more then one B mode, find the one with largest chmax

  foreach $m (@B)
  {
    ($mode, $id, $binsize, $chnum, $chmin, $chmax) = @$m;
    if($chmax > $max) {$max = $chmax; $m_max = $m};
  }
  if(@B > 0) {push @PDS, [@$m_max];}

#  if(@B > 1) {die "I'm confused, more than one Binned mode found.\n"}
#  if(@B == 1) {push @PDS, $B[0];}

  return @PDS;

}
    


1;
