package Xray::Readopt;
require Exporter;

use strict;
use warnings;

use Term::Complete;
use Getopt::Long;

our @ISA = qw(Exporter);
our @EXPORT = qw(ReadOptions);
our $VERSION = 1.0;


sub ReadOptions
#
# Read options from the command line and, if not provided, ask for them.
#
# my %opt = ReadOptions($rcfile, \%optrec);
#
# $rcfile - history file in which previous options are stored
# %optrec = (
#   option1 => [1, 's', 'Give option 1 (string)'],
#   option2 => [2, 'f', 'Give option 2 (real number)'],
#   option3 => [-1, 's', 'Command-line only option, do not ask']
# )
#
# Returns a hash of option_name => value.
#
{
  my ($file, $optrec) = @_;
  
  local *F;

# read saved values from the file
    
  my %sav = ();
  if(open F, $file)
  {
    while(my $line = <F>)
    {
      chomp $line;
      if($line =~ /(\w+)\s*=\s*(\S+)/)
        {$sav{$1} = $2;}
    }
  }
  
# read options from command line
  
  my @par = ('help', 'man');  # always allowed
  for my $x (keys %$optrec)
  {
    my $type = $optrec->{$x}->[1];
    if($type =~ /[sif]/)
      {$type = "=$type";}
    elsif($type ne '')
    {
      print "\n$0: unrecognized type \'$type\' in option \'$x\', no argument assumed\n";
      $type = '';
    }
    push @par, "${x}${type}";
  }
  my %opt;
  GetOptions(\%opt, @par);
  
  return %opt if $opt{help} || $opt{man};

# ask for undefined options
  
  local *DIR;
  opendir(DIR, '.') or die "$!\nSerious drainbamage.\n";
  my @d = readdir(DIR);
  close(DIR);
  for my $optname (sort {$optrec->{$a}->[0] cmp $optrec->{$b}->[0]} keys %$optrec)
  {
    next if defined $opt{$optname} || $optrec->{$optname}->[1] eq '' || $optrec->{$optname}->[0] < 0;
    
    my $def = (defined $sav{$optname}) ? $sav{$optname} : '';
    my $descr = $optrec->{$optname}->[2];

    my $answer = Complete("$descr [$def]: ", \@d);
    if($answer eq '') {$answer = $def;}
    if($answer eq 'kill') {die "\nKilled.\n";}
    $opt{$optname} = $answer;
  }

# save options to the file
    
  if(open F, ">$file")
  {
    for my $optname (keys %opt)
      {print F "$optname = $opt{$optname}\n";}
  }
  
  return %opt;
}

1;

__END__

=head1 NAME

B<ReadOptions> - read command-line options and/or ask for them

=head1 SYNOPSIS

  use Xray::Readopt;
  my $rcfile = ".rcfile";
  my %optrec = (
    file =>    [1, 's', 'Input file'],          # 1st argument, string
    length =>  [2, 'i', 'Length'],              # 2nd argument, numeric
    alter =>   [-1, 's', 'Alternate method'],     # hidden argument, string
  );
  my %opt = ReadOptions($rcfile, \%optrec);
  
  my $filename = $opt{file};
  my $intlength = $opt{length};
  print "Using alternate method $opt{alter}\n" if $opt{alter};
  print "This is help\n" if $opt{help};
  print "Manual page\n" if $opt{man};
  print "File $filename, length $intlength\n";

=head1 DESCRIPTION
  
Reads command-line options in a traditional format used by L<Getopt::Long>, e.g., C<-file=test.dat> or C<--file=test.dat>. If an option is defined in C<%optrec> but not provided in the command line, user will be asked for the value of this option, using the string provided. Each value is remembered in the history file, C<$rcfile>, and used as a default next time the code is invoked.

In the example given above the Perl script F<test.pl> containing these lines can be invoked:

  > test.pl -file=test.dat -length=23
  File test.dat, length 23
  
When no command-line options are given the code will ask the user:

  > test.pl
  Input file []: test.dat
  Length []: 23
  File test.dat, length 23
  
When the code is invoked again, it will read previous options from the history file F<$rcfile> and use them as defaults. To accept the default, hit enter.
  
  > test.pl
  Input file [test.dat]:
  Length [23]: 30
  File test.dat, length 30
  
The order of questions is defined by numbers given in anonymous lists in the hash C<%optrec>. However, if the number is negative, the option is "hidden", ie. the code will not ask for this even if it is not given in the command line. Hence:

  > test.pl -alter=new
  Input file [test.dat]:
  Length [30]:
  Using alternate method new
  File test.dat, length 30

