package Xray::Statistics;
require Exporter;

use strict;
use warnings;

use Math::Trig;

our $VERSION = 1.00;

our @ISA = ("Exporter");
our @EXPORT = qw(
 Gammaln
 Gauss
 MeanStddev
 Poisson
);

my $PI = 4 * atan(1);

################################################################

sub Gammaln
#
#  Log of gamma function, ln(Gamma(z)), z > 0
#  (from Numerical Recipes)
{
  my $z = shift;
  
  my @cof = (76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5);
  
  my $x = $z - 1;
  my $tmp = $x + 5.5;
  $tmp = ($x + 0.5)*log($tmp) - $tmp;
  my $ser = 1;
  for my $c (@cof)
  {
    $x += 1;
    $ser += $c / $x;
  }
  return $tmp + log(2.5066282746310005 * $ser);
}

################################################################

sub Gauss
#
# Box-Muller Gaussian generator
#
# $g = Gauss($mean, $sigma);
#
{
  my ($mean, $sigma) = @_;
  my ($x1, $x2, $w, $y1);
  
  do{{
    $x1 = 2 * rand() - 1;
    $x2 = 2 * rand() - 1;
    $w = $x1 * $x1 + $x2 * $x2;
  }} while($w >= 1);
  $w = sqrt((-2 * log($w))/$w);
  $y1 = $x1 * $w;
  
  $y1 = $y1 * $sigma + $mean;
  return($y1);
}

################################################################

sub MeanStddev
#
#  ($mean, $sigma) = MeanStddev(\@lc);
#
# Finds mean and standard deviation for a given array of data,
# using a canning little recursive algorithm allowing for one loop and
# avoiding potential dangers of rounding up on subtraction and
# producing negative variance.
{
  my $dat = shift;
  
  my $m = $dat->[0];
  my $s = 0;
  my $p = 2;
  my $first = 1;
  for my $x (@$dat)
  {
    if($first) {$first = 0; next;}
    $m = (($p - 1)*$m + $x)/$p;
    $s = (($p - 2)*$s + $p/($p - 1) * ($x - $m)**2)/($p - 1);
    $p++;
  }
  $s = sqrt($s);
  return($m, $s);
}

################################################

BEGIN {

my $oldm = -1;         # 'static' variables
my ($sq, $alxm, $g);

sub Poisson
#
# Generates a number with a Poissonian distribution.
#
# $x = Poisson($mean);
#
#  (from Numerical Recipes)
{
  my $xm = shift;
  
  my $em;
  
  if($xm < 12)
  {
    if($xm != $oldm)
    {
      $oldm = $xm;
      $g = exp(-$xm);
    }
    $em = -1;
    my $t = 1;
    do {
      $em++;
      $t *= rand();
    } while($t > $g);
  }
  
  else  # $xm >= 12
  
  {
    if($xm != $oldm)
    {
      $oldm = $xm;
      $sq = sqrt(2 * $xm);
      $alxm = log($xm);
      $g = $xm * $alxm - Gammaln($xm + 1);
    }
    my $t;
    do {
      my $y;
      do {
        $y = tan($PI * rand());
        $em = $sq * $y + $xm;
      } while($em < 0);
      $em = int($em);
      $t = 0.9 * (1 + $y*$y) * exp($em * $alxm - Gammaln($em + 1) - $g);
    } while(rand() > $t);
  }
  return $em;
}

}


1;
