package Xray::xspec;
require Exporter;

use strict;
use warnings;

use PDL;
use PDL::Graphics::PGPLOT;
use PDL::NiceSlice;
use Xray::Graphics;
use Xray::Tools;

our @ISA = qw(Exporter);
our @EXPORT = qw(
  PlotSpec
  PlotFitCont
  $xspec
  $pg_lwidth
  $pg_unabs
  $pg_erg9
  $pg_corfree
  @pg_linecolours
  @pg_linestyles
  @pg_datcolours
);

our $xspec = 'xspec';  # call to xspec version 11, needs an alias!

our $pg_lwidth = 2;
our $pg_unabs = 0;
our $pg_erg9 = 0;
our $pg_corfree = 1;  # correct free normalization by default

our @pg_linecolours = (1 .. 128);
our @pg_linestyles = (1) x 128;
our @pg_datcolours = (1) x 128;



sub PlotSpec
#
# PlotSpec($name)
#
# Plots unfolded spectrum with model components. $name is the root
# for files: $name_da.qdp $name_mo.qdp $name_mona.qdp and/or $name_monr.qdp
# 
{
  my ($name) = @_;
  
  my $data = "${name}_da.qdp";
  my $model = "${name}_mo.qdp";
  my $unabs = "${name}_mona.qdp";
  my $noref = "${name}_monr.qdp";
  my $is_abs = (-e $unabs);
  my $is_ref = (-e $noref);
  my $is_dat = (-e $data);
  
  my $erg9 = ($pg_erg9) ? 1.602185 : 1;
  
  Rmmin($data) if($is_dat);
  Rmmin($model);
  Rmmin($unabs) if $is_abs;
  Rmmin($noref) if $is_ref;
  
  my @spec = rcols $model, {Lines => '3:-1'};
  my @unspec = rcols $unabs, {Lines => '3:-1'} if($is_abs);
  my $engy = $spec[0];
  my $tot = $spec[2] * $erg9 + 1e-20;  # infinity trap
  my $untot = ($is_abs) ? $unspec[2] * $erg9 : $tot;
  my $abs = $untot / $tot;
  my $lengy = nlog10($engy);

  my $lref;
  if($is_ref)
  {
    my $nr = rcols $noref, 2, {Lines => '3:-1'};
    my $ref = $untot - $nr * $erg9;
    if(!$pg_unabs)
      {$ref = $ref / $abs;}
    $lref = nlog10($ref);
  }
  
  my @plot = ($pg_unabs && $is_abs) ? @unspec : @spec;
  shift @plot; shift @plot;
  
  if($is_dat)
  {
    my $datnum = 0;
    my $p1 = 3;
    my $p2;
    my @no = `grep -n NO $data`;
    for my $n (@no)
    {
      my ($pos) = split /:/, $n;    
      $p2 = $pos - 2;

      pdat($data, $p1, $p2, $engy, $abs, $tot, $datnum);
      $datnum++;
      $p1 = $p2 + 2;
    }
    pdat($data, $p1, '-1', $engy, $abs, $tot, $datnum);
  }

  my $num = 0;
  foreach my $comp (@plot)
  {
    my $curve = nlog10($comp * $erg9);
    line $lengy, $curve, {Colour => $pg_linecolours[$num], LineStyle => $pg_linestyles[$num], LineWidth => $pg_lwidth};
    $num++;
  }
  if($is_ref)
  {
    line $lengy, $lref, {Colour => $pg_linecolours[$num], LineStyle => $pg_linestyles[$num], LineWidth => $pg_lwidth};
  }
    
}

sub pdat
{
  my ($data, $p1, $p2, $engy, $abs, $tot, $num) = @_;
  
  my $erg9 = ($pg_erg9) ? 1.602185 : 1;
  
  my ($x, $dx, $y, $dy, $mody) = rcols $data, {Lines => "$p1:$p2"};
  my $n = at(nelem($x),0);
  print "  Read $n lines $p1:$p2\n";
  
  $y *= $erg9;
  $dy *= $erg9;
  $mody *= $erg9;

  # correction for free normalization of a data set
  
  my $corr = 1;
  if($pg_corfree)
  {
    my $i = 0;
    my $j = 0;
    while($j <= $n - 1)
    {
      while(at($engy, $i) < at($x, $j)) {$i++;}
      last if (at($abs, $i) < 3 && at($mody, $j) != 0);   # avoid highly absorbed points
      $j++;
    }
    $j-- if($j == $n);
    my $m = at($mody, $j);
    $corr = at($tot, $i) / $m unless $m == 0;
  }
  
  unabsorb($x, $y, $dy, $engy, $abs) if($pg_unabs);
 
  my ($lx, $x_l, $x_u) = LogSerr($x, $dx);
  my ($ly, $y_l, $y_u) = LogSerr($y * $corr, $dy * $corr);
  errb $lx, $ly, $x_l, $x_u, $y_l, $y_u, {Colour => $pg_datcolours[$num], Charsize => 1, Term => 0, Symbol => $pg_sym{dot}, LineWidth => $pg_lwidth};
#  wcols $lx, $ly, $x_l, $x_u, $y_l, $y_u;
}

sub unabsorb
{
  my ($x, $y, $dy, $e, $abs) = @_;
 
  my $rabs = zeroes $x;
  my $j = 0;
  for my $i (0 .. nelem($x) - 1)
  {
    #DO: do{
    #  $j++;
    #  my $ex = at($x, $i);
    #  die "\nEnergy aray does not cover data at E = $ex keV\n" if($j >= nelem($e));
    #  last DO if(at($e, $j) >= at($x, $i));
    #};
    while(at($e, $j) < at($x, $i)) 
    {
      $j++;
      my $ex = at($x, $i);
      die "\nEnergy grid in model does not cover data at E = $ex keV. Cannot unabsorb.\n" if($j >= nelem($e));
    }
    #    $rabs($i) .= $abs($j);
    $rabs($i) .= $abs($j-1) + ($abs($j) - $abs($j-1)) * ($x($i) - $e($j-1)) / ($e($j) - $e($j-1));
    
    my $xx = at($x, $i);
    my $ee = at($e, $j);
    my $rr = at($rabs, $i);
#    print "$i $j   $xx  $ee    $rr\n";
  }
  $y *= $rabs;
  $dy *= $rabs;
}

sub PlotFitCont
#
# Create contour plot from steppar output. After steppar had been
# run the chi^2 matrix you need to save it (ipl; wda file) and
# remove hyphenation (rmmin). This can be used for contour plot.
# 
# PlotFitCont($file,
# {
#  box => [$minx, $maxx, $miny, $maxy],
#  bestx => $best_fitting_x,
#  besty => $best_fitting_y,
#  minchi => $best_fitting_chi^2,
#  levels => [$lev1, $lev2, ..., $levn]
# });
#
# minchi and levels are not compulsory. Levels are delta chi^2 levels.
# if not specified, [2.30, 4.61, 9.21] are used (68%, 90% and 99%).
{
  my ($file, $opt) = @_;

  my $levels = (defined $opt->{levels}) ? 
    pdl($opt->{levels}) : pdl(2.30, 4.61, 9.21);  # 68%, 90%, 99%
  my $img = rmatrix($file);
  my ($nx, $ny) = dims($img);
  my ($xmin, $xmax, $ymin, $ymax) = (defined $opt->{box}) ?
    @{$opt->{box}} : (0, $nx - 1, 0, $ny - 1);
  my $xrate = ($xmax - $xmin) / ($nx - 1);
  my $yrate = ($ymax - $ymin) / ($ny - 1);
  my $bestx = (defined $opt->{bestx}) ? 
    $opt->{bestx} : $xmin + ($xmax - $xmin) / 2;
  my $besty = (defined $opt->{besty}) ? 
    $opt->{besty} : $ymin + ($ymax - $ymin) / 2;
  my ($minchi, $maxchi);
  if(!defined $opt->{minchi})
    {($minchi, $maxchi) = minmax($img);}
  else
    {$minchi = $opt->{minchi};}
    
  my $tr = transform(dims($img), {ImageCenter => [$bestx, $besty], Pixinc => [$xrate, $yrate]});
  my $conts = $minchi + $levels;
  cont $img, {Contours => $conts, Transform => $tr, Colour => 'Black'};
  points pdl($bestx), pdl($besty), {Colour => 'Red', Symbol => $pg_sym{x}, Charsize => 1.1};
}


1;
